//************************************************************************
// 	HIERBOX.CPP
//************************************************************************
//
//  Purpose:	Hierarchical listbox
//
//  Author:	ABRI Datenverarbeitung GesmbH
//		Andreas Vogel
//		Miltnerweg 17/14/19
//		A-1110 Vienna, Austria
//		phone/fax/modem: 0043 1 76 23 35
//		handy:		 0043 0663 802 538
//		CIS:		 100343,2704
//
//  History:
//
//    	94-07-25  V1.00	created from MSNCD V7.0
//
//	--------------------------------------------------------
//
//  Brief description:
//
//	Implementation of a Hierarchical ListBox using bitmaps in the
//	front of the item text.
//	  Folding can be either done by double-clicking on an item or
//	by single-clicking on the bitmap. Further the single-click is
//	also simulated by pressing the BLANK-key on the selected item.
//
//	Constructor1:
//
//	  TWindow * Parent	parent window (normally the dialog)
//	  int       Id		ID of listbox
//	  int	    BitmapId	id of bitmap (see below)
//	  int	    rows	# of rows in bitmap
//	  int	    columns     # of columns in bitmap
//	  int	    loins	TRUE == draw connection lines
//	  TModule * modul	module where to load resources (bitmap)
//
//	Constructor2:
//
//	  TWindow * Parent	parent window (normally the dialog)
//	  int       Id		ID of listbox
//	  char * DefaultBitmap  bitmap to adjust height of a line
//	  int	    loins	TRUE == draw connection lines
//	  TModule * modul	module where to load resources (bitmap)
//

#include "hierbox.h"

//-------------------------------------------------------- defines -------
//-------------------------------------------------------- globals -------
//------------------------------------------------------ externals -------
//--------------------------------------------- external functions -------
//------------------------------------------------ local functions -------
//-------------------------------------------------------- statics -------

#define WINDOWBKGNDCOLOR (GetSysColor(COLOR_WINDOW))

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
// 	class cHierListBox
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

DEFINE_RESPONSE_TABLE1(cHierListBox, TListBox)
  EV_WM_DRAWITEM,
  EV_WM_MEASUREITEM,
  EV_WM_LBUTTONDOWN,
  EV_WM_CHAR,
END_RESPONSE_TABLE;

//------------------------------------------------------------------------
	cHierListBox::cHierListBox(TWindow * Parent,
				 int       Id,
				 int	   BitmapId,
				 int	   rows,
				 int	   columns,
				 int	   loins,
				 TModule * modul) :

		TListBox(Parent, Id, modul),
		Content(0, 0, 5)
//------------------------------------------------------------------------
// Approach 1:	Bitmaps are stored in a single Bitmap in the resource file
{
  HRSRC		hRes;
  HGLOBAL	hResMem;

  LPBITMAPINFOHEADER lpbi;

  DWORD FAR * 	lpColorTable;
  LPSTR 	lpBits;
  int 		bc;
  HDC 		hDC;

  Ok = 0;
  XLinePos = 2;		// default inserting is 50% of symbol

  nBitmapHeight = 0;  	// default value for the case of error
  nBitmapWidth  = 8;  	// default value for the case of error

  bLines = loins;

  nBitmapRows = rows;
  nBitmapColumns = columns;

  // calculate char height & item height
  // take the font from the dialog:
  TEXTMETRIC   TextMetrics;
  TWindowDC    hdc(HWindow);

  hdc.GetTextMetrics(TextMetrics);

  nLineHeight = nTextHeight = TextMetrics.tmHeight;


  hDC = GetDC(NULL);
  hdcMem = CreateCompatibleDC(hDC);
  ReleaseDC(NULL,hDC);

  if (!hdcMem)
    return;

  hbmMem = NULL;

  // Load the Bitmap ( original from disk )

  hRes = GetModule()->FindResource(MAKEINTRESOURCE(BitmapId), RT_BITMAP);

  if (!hRes)
      return;

  hResMem = GetModule()->LoadResource(hRes);

  if (!hResMem)
      return;

  // Now figure out the bitmaps background color.
  // This code assumes the these are 16 color bitmaps
  // and that the lower left corner is a bit in the background
  // color.

  lpbi = (LPBITMAPINFOHEADER)LockResource(hResMem);

  if (!lpbi)
      return;

  lpColorTable = (DWORD FAR *)(lpbi + 1);

  lpBits = (LPSTR)(lpColorTable + 16);            // ASSUMES 16 COLOR

  bc = (lpBits[0] & 0xF0) >> 4;                   // ASSUMES 16 COLOR
			  // ALSO ASSUMES LOWER LEFT CORNER IS BG!!!

  lpColorTable[bc] = RGB2BGR(WINDOWBKGNDCOLOR);

  hDC = GetDC(NULL);

  hbmIcons = CreateDIBitmap(hDC,
			    lpbi,
			    (DWORD) CBM_INIT,
			    lpBits,
			    (LPBITMAPINFO) lpbi,
			    DIB_RGB_COLORS);
  ReleaseDC(NULL,hDC);

  nBitmapHeight = (WORD)lpbi->biHeight / rows;
  nBitmapWidth  = (WORD)lpbi->biWidth / columns;

  nLineHeight = max(nBitmapHeight, nTextHeight);

  UnlockResource(hResMem);
  FreeResource(hResMem);

  if (!hbmIcons)
    return;

  hbmMem = (HBITMAP) SelectObject(hdcMem, hbmIcons);

  if (!hbmMem)
    return;

  Ok = 1;

  return;
}

//------------------------------------------------------------------------
      cHierListBox::cHierListBox(TWindow * Parent,
				 int       Id,
				 char *	   DefaultBitmap,
				 int	   loins,
				 TModule * modul) :

		TListBox(Parent, Id, modul),
		Content(0, 0, 5)
//------------------------------------------------------------------------
// Approach 2:	Bitmaps are externally stored specific Bitmap files
//		The assignment is performed by the AddItem() function
//
// The <DefaultBitmap> adjusts the height of a line and is always used
// when a specified file in AddItem() cannot be opened. The background
// of the Bitmap is not out-masked as it is done by the first constructor.
{
  LPBITMAPINFOHEADER lpbi;

  HDC 		hDC;

  bLines = loins;

  Ok = 0;
  XLinePos = 2;		// default inserting is 50% of symbol

  nBitmapHeight = 0;  	// default value for the case of error
  nBitmapWidth  = 8;  	// default value for the case of error

  nBitmapRows    = 1;
  nBitmapColumns = 1;

  // calculate char height & item height
  // take the font from the dialog:
  TEXTMETRIC   TextMetrics;
  TWindowDC    hdc(HWindow);

  hdc.GetTextMetrics(TextMetrics);

  nLineHeight = nTextHeight = TextMetrics.tmHeight;


  hDC = GetDC(NULL);
  hdcMem = CreateCompatibleDC(hDC);
  ReleaseDC(NULL,hDC);

  if (!hdcMem)
    return;

  TDib *dib = NULL;

  try
  {
    dib = new TDib(DefaultBitmap);	// Load the default Bitmap from disk
  }

  catch (TGdiBase::TXGdi)
  {
    Parent->MessageBox("Cant load Default-Bitmap",
		       GetApplication()->GetName(),
		       MB_OK);
    return;
  }

  if (!dib->IsOK())		// check if loading fails
    return;

  lpbi = dib->GetInfoHeader();	// get BITMAPINFOHEADER

  if (!lpbi)
    return;

  hDC = GetDC(NULL);

  hbmIcons = CreateDIBitmap(hDC,
			    lpbi,
			    CBM_INIT,
			    dib->GetBits(),
			    (LPBITMAPINFO) lpbi,
			    DIB_RGB_COLORS);

  ReleaseDC(NULL,hDC);

  nBitmapHeight = (WORD)lpbi->biHeight;
  nBitmapWidth  = (WORD)lpbi->biWidth;

  nLineHeight = max(nBitmapHeight, nTextHeight);

  if (!hbmIcons)
    return;

  hbmMem = (HBITMAP) SelectObject(hdcMem, hbmIcons);

  if (!hbmMem)
    return;

  Ok = 1;

  return;
}

//------------------------------------------------------------------------
	cHierListBox::~cHierListBox()
//------------------------------------------------------------------------
{
  if (hbmMem)
    SelectObject(hdcMem, hbmMem);	// restore old memory object
  if (hdcMem)
    DeleteDC(hdcMem);			// remove memory DC

  // remove the memory bitmaps for the approach2
  for (int i=0; i<Content.GetItemsInContainer(); i++)
    if (Content[i].hbm)
      DeleteObject(Content[i].hbm);
}

//------------------------------------------------------------------------
void	cHierListBox::SetupWindow()
//------------------------------------------------------------------------
{
  SetBkgndColor(WINDOWBKGNDCOLOR);
}

//------------------------------------------------------------------------
void	cHierListBox::EvDrawItem(UINT 		     /*ctrlId*/,
				DRAWITEMSTRUCT far & draw)
//------------------------------------------------------------------------
{
  DWORD dwData;
  int   nLevel;
  int   nRow;
  int   nColumn;
  DWORD dwConnectLevel = 0L;
  DWORD dwMask;

  dwData = draw.itemData;

  int	index = GetIndex(dwData); 	// index to container data

  // Select the correct icon, open folder, closed folder, or document

  if (Foldable(index) )  		// Can this item be opened ?
  {
    if ( IsOpened(dwData) )		// is it currently open ?
      nColumn = 1;
    else
      nColumn = 0;
  }
  else
    nColumn = 2;

  // Select the correct level/color
  // This sample changes the row ( color ) just because it can.
  // restart with the fst color on limit

  nLevel = GetLevel(index);
  nRow   = GetLevel(index) % nBitmapRows;

  // Figure out which connecting lines are needed.
  // If this item is the last kid or one it parents
  // is a last kid, then it does not need a connector at that level

  if ( nLevel == 0 )
  {
    dwConnectLevel = 0L;	// Level 0 items never have connectors
  }
  else
  {
    int	idx = index;			// start searching at current item

    dwMask = 1L << (nLevel - 1);	// start with deepest value

    // build connectors_lines mask, level 0 has NO lines

    for (int i=nLevel; i>0; i--, dwMask>>=1)
    {
      if ( !LastChild(idx) )
	dwConnectLevel |= dwMask;

      idx = PrevParent(idx);	// go back one level
    }
  }

  // call drawing function
  OnDrawItem( draw,
	      nLevel,
	      dwConnectLevel,
	      Content[index].Item,
	      Content[index].hbm,
	      nRow,
	      nColumn);
}

//------------------------------------------------------------------------
void		cHierListBox::EvMeasureItem(UINT /*ctrlId*/,
					    MEASUREITEMSTRUCT far & draw)
//------------------------------------------------------------------------
{
  draw.itemHeight = nLineHeight;
}

//------------------------------------------------------------------------
void	cHierListBox::EvLButtonDown(UINT u, TPoint &p)
//------------------------------------------------------------------------
// support foldup/down with pressing a single mouse-click on the bitmap
{
  int 	nLevel;
  int	sidx;
  TRect rcTemp;

  TListBox::EvLButtonDown(u, p);

  sidx = GetSelIndex();		// get selected item

  if (sidx < 0)
    return;

  GetClientRect(rcTemp);

  nLevel = GetLevel(GetItemData(sidx));		// get level of selected entry

  // calculate range where to click with mouse
  rcTemp.left  +=  ((int)(nLevel) * 2*nBitmapWidth/XLinePos) + XBMPOFFSET;
  rcTemp.right  = rcTemp.left + nBitmapWidth;
  rcTemp.top    = rcTemp.top + nLineHeight * (sidx - GetTopIndex());
  rcTemp.bottom = rcTemp.top + nLineHeight;

  // when mouse is pressed on bitmap ==> action
  if (p.x > rcTemp.left && p.x < rcTemp.right &&
      p.y > rcTemp.top  && p.y < rcTemp.bottom )
    Action();
}

//------------------------------------------------------------------------
void	cHierListBox::EvChar(UINT key, UINT r, UINT f)
//------------------------------------------------------------------------
{
  TListBox::EvChar(key, r, f);

  if (key == ' ')
    Action();
}

//------------------------------------------------------------------------
BOOL 	cHierListBox::Action()
//------------------------------------------------------------------------
//	action for double click on item
//
// the user must call this function from the dialog procedure to setup
// the folding process. When last foldup has performed, TRUE is returned.
{
  WORD  wItemNum = GetSelIndex();
  DWORD	dwData   = GetItemData(wItemNum);

  int	index = GetIndex(dwData);	// index to <Content>

  int	sellevel = GetLevel(index);	// level of current item

  if ( !Foldable(index) )
    return 1;			// item is not foldable --> action
  else
  {
    if ( IsOpened(dwData) )
    {
      CloseItem(wItemNum);	   // remove "open" information from <Content>

      wItemNum++;

      while (1)
      {
	if (GetCount() <= wItemNum)	// check on last item in listbox
	  break;

	if (GetLevel(GetItemData(wItemNum)) > sellevel)
	{
	  CloseItem(wItemNum);	   // remove "open" information from <Content>

	  DeleteString(wItemNum);  // remove from listbox
	}
	else
	  break;
      }
    }
    else
    {
      int 	i 	= wItemNum + 1;
      int	NewKids = 0;		// # of new items appended

      index++;	// next index in <Content> array

      OpenItem(wItemNum);	// set "open" information in <Content>

      SendMessage(WM_SETREDRAW, FALSE, 0L);   // Disable redrawing

      for (; index<Content.GetItemsInContainer(); index++)
      {
	if (GetLevel(index) == sellevel + 1)
	{
	  NewKids++;

	  InsertString(Content[index].Item, i);

	  SetItemData(i, Content[index].Data);

	  i++;		// next index in Listbox
	}

	if (GetLevel(index) <= sellevel)
	  break;
      }

      SendMessage(WM_SETREDRAW, TRUE, 0L);   // Enable redrawing.
      Invalidate();		              // Force redraw

      // Make sure as many child items as possible are showing
      ShowKids(wItemNum, NewKids);
    }
  }
  return 0;
}

//------------------------------------------------------------------------
void	cHierListBox::AddItem(const char * str,
			     const int 	  level,
			     const BOOL   foldup,
			     const BOOL   selectable)
//------------------------------------------------------------------------
// Append an item to the internal list.
// The order the items are created is VERY important, since this information
// is used as decision if there are level jumps
//
// str:		Item String
//
// level:	level to position the string in the listbox
//
// foldup:	TRUE = 	foldup this item (works only when higher level
//			is following). ALL items which are child of this
//			item are folded up !
//
// selectable:	TRUE = 	this item returns TRUE when item is double clicked
//			also a foldable item can be made selectable
//
{
  DWORD dw = (level << 8) | (selectable ? 1 : 0);

  // setup index <I> field for ListBox-access to Content-access
  dw |= ((DWORD)Content.GetItemsInContainer()) << 16;

  if (foldup)
    dw |= 0x80000000L;

  Content.Add( HierListBoxData(str, dw) );
}

//------------------------------------------------------------------------
void	cHierListBox::AddItem(const char * BitmapFile,
			      const char * str,
			      const int    level,
			      const BOOL   foldup,
			      const BOOL   selectable)
//------------------------------------------------------------------------
// Append an item to the internal list.
// The order the items are created is VERY important, since this information
// is used as decision if there are level jumps
//
// BitmapFile:	Filename of external bitmap
//
// str:		Item String
//
// level:	level to position the string in the listbox
//
// foldup:	TRUE = 	foldup this item (works only when higher level
//			is following). ALL items which are child of this
//			item are folded up !
//
// selectable:	TRUE = 	this item returns TRUE when item is double clicked
//			also a foldable item can be made selectable
//
{
  DWORD dw = (level << 8) | (selectable ? 1 : 0);

  // setup index <I> field for ListBox-access to Content-access
  dw |= ((DWORD)Content.GetItemsInContainer()) << 16;

  if (foldup)
    dw |= 0x80000000L;

  HBITMAP 	hbm = 0;
  TDib *	dib = NULL;

  try
  {
    dib = new TDib(BitmapFile);	// Load the Bitmap from disk
  }

  catch(TGdiObject::TXGdi& x)
  {
    // append only text to container
    Content.Add( HierListBoxData(str, dw) );
    return;
  }

  if (dib->IsOK())		// check if loading fails
  {
    HDC hDC = GetDC(NULL);

    hbm = CreateDIBitmap(hDC,
			 dib->GetInfoHeader(),
			 CBM_INIT,
			 dib->GetBits(),
			 (LPBITMAPINFO) dib->GetInfoHeader(),
			 DIB_RGB_COLORS);

    ReleaseDC(NULL, hDC);
  }

  Content.Add( HierListBoxData(str, dw, hbm) );
}

//------------------------------------------------------------------------
void	cHierListBox::SetupItemData(BOOL foldup)
//------------------------------------------------------------------------
// setup the flags in the data-item and append the main entries
// to the listbox
//
// when <foldup> is TRUE, ALL items are folded up. Otherwise the information
// from the data-word is taken to decide whether the item should be folded up
// or not
{
  HierListBoxData *	LastItem;		// last item
  int			PrevLevel = 0;     	// last level
  int			j = 0;
  int			FoldUpLevel = 1000;	// holds information bout level
						// to fold up

  int			Items = Content.GetItemsInContainer();

  ClearList();	// remove all items from array

  LastItem  = &Content[0];

  for (int i=1; i<Items; i++)
  {
    int lev = (int) (Content[i].Data & 0xFF00) >> 8;	// level

    LastItem->Data &= ~0x06;	// remove "last item" & "foldable" flag

    if ( lev < PrevLevel)
    {
      // lower level ==> prev. entry is last in chain
      LastItem->Data |= 0x04;
    }
    else
      if ( lev > PrevLevel)
      {
	// higher level ==> prev. entry is foldable
	LastItem->Data |= 0x02;

	if (LastItem->Data & 0x80000000L)
	  FoldUpLevel = PrevLevel;   	// fold up all items higher this level

	// search forward, if the previous foldable entry is the
	// last one in the previous list
	for (int j=i; j<Items; j++)
	{
	  int l = (int) (Content[j].Data & 0xFF00) >> 8; // level

	  if (l < PrevLevel)
	    LastItem->Data |= 0x4;

	  if (l == PrevLevel)	// abort searching when same level has found,
	    break;		// cause then there exists a follower
	}

	// last item in the chain with the same level
	if (j==Items)
	  LastItem->Data |= 0x4;
      }
      else      // same level
      {
      }

    if ( !(LastItem->Data & 0x02) )
    {
      // when item is not foldable it IS selectable !
      LastItem->Data |= 1;
    }

    if (!GetLevel(i-1) 		  ||	// start with all items closed show only level 0
	foldup		    	  ||	// or fold up all
	FoldUpLevel<GetLevel(i-1)  )    // or foldup all childs on a level
    {
      if ( LastItem->Data & 0x02 )
      {
	if ( FoldUpLevel<GetLevel(i-1) )
	{
	  // set "folded up" for deeper values
	  LastItem->Data |=  0x80000000L;
	}
      }

      if (lev == FoldUpLevel)
	FoldUpLevel = 1000;		// nothing more to fold up

      AddString(LastItem->Item);	// append the previous item
      SetItemData(j++, LastItem->Data);
    }

    PrevLevel = lev;
    LastItem  = &Content[i];
  }

  // the last item MUST always be the last in chain
  LastItem->Data |= 0x04;

  // append the last item to the listbox
  if (!GetLevel(i-1) || foldup)
  {
    LastItem->Data |= 0x04;		// is always the last in chain

    AddString(LastItem->Item);
    SetItemData(j++, LastItem->Data);
  }
}

//------------------------------------------------------------------------
void	cHierListBox::OpenItem(int listindex)
//------------------------------------------------------------------------
// <listindex> is the index to the listbox NOT to the <Content> array
{
  DWORD dw = GetItemData(listindex);	// get the data field

  int	idx = GetIndex(dw);		// convert to <Content> index

  Content[idx].Data |= 0x80000000L;     // set the "shown" flag
}

//------------------------------------------------------------------------
void	cHierListBox::CloseItem(int listindex)
//------------------------------------------------------------------------
// <listindex> is the index to the listbox NOT to the <Content> array
{
  DWORD dw = GetItemData(listindex);	// get the data field

  int	idx = GetIndex(dw);		// convert to <Content> index

  Content[idx].Data &= ~0x80000000L;	// reset the "shown" flag
}

//------------------------------------------------------------------------
int	cHierListBox::Foldable(int idx)
//------------------------------------------------------------------------
// check if item is foldable
{
  return (Content[idx].Data & 0x2) != 0;
}

//------------------------------------------------------------------------
int	cHierListBox::LastChild(int idx)
//------------------------------------------------------------------------
// check if item is the last in the chain
{
  return (Content[idx].Data & 0x4) != 0;
}

//------------------------------------------------------------------------
int	cHierListBox::PrevParent(int idx)
//------------------------------------------------------------------------
// get the prev. parent item
{
  int 	lv = (int) (Content[idx].Data & 0xFF00);

  for (int i=idx; i>=0; i--)
    if (lv > (Content[i].Data & 0xFF00))
      return i;

  return 0;
}

//------------------------------------------------------------------------
void	cHierListBox::ShowKids(WORD wCurrentSelection,
			      WORD wKids)
//------------------------------------------------------------------------
// adjust top-index of listbox to show as many childs as possible
{
  WORD 	wBottomIndex;
  WORD 	wTopIndex;
  WORD 	wNewTopIndex;
  WORD 	wExpandInView;
  TRect rc;

  wTopIndex = GetTopIndex();	// get current top-most index

  GetClientRect(rc);		// get size of listbox

  // max. index which can be shown with the top index
  wBottomIndex = wTopIndex + (rc.bottom+1) / nLineHeight;

  wExpandInView = (wBottomIndex - wCurrentSelection);

  if (wKids >= wExpandInView)
  {
      wNewTopIndex = min((unsigned) wCurrentSelection,
			 (unsigned) wTopIndex + wKids - wExpandInView + 1);

      SetTopIndex(wNewTopIndex);
  }
}

//------------------------------------------------------------------------
void 	cHierListBox::OnDrawItem(DRAWITEMSTRUCT far & draw,
				int   		     nLevel,
				DWORD 		     dwConnectLevel,
				char *		     szText,
				HBITMAP		     hbm,
				int   		     nRow,
				int   		     nColumn)
//------------------------------------------------------------------------
// paint single item
{
  WORD       wIndent, wTopBitmap, wTopText;
  TRect      rcTemp;
  char	     *BreakStr;

  if ( draw.itemID == (UINT)-1 )
     return ;

  rcTemp = draw.rcItem;

  wIndent 	= rcTemp.left +
		  ((int)(nLevel) * 2*(nBitmapWidth/XLinePos)) +
		  XBMPOFFSET;

  rcTemp.left 	= wIndent + nBitmapWidth;

  wTopText 	= rcTemp.top +
		  ((rcTemp.bottom - rcTemp.top) / 2) -
		  (nTextHeight / 2);

  wTopBitmap 	= rcTemp.top +
		  ((rcTemp.bottom - rcTemp.top) / 2) -
		  (nBitmapHeight / 2);

  if (draw.itemAction == ODA_FOCUS)
  {
    ::DrawFocusRect(draw.hDC, &rcTemp);
    return;
  }
  else
    if (draw.itemAction == ODA_SELECT)
    {
      goto DealWithSelection;
    }

  // Draw some lions, if we like lions

  if (bLines && nLevel)
  {
    DWORD    dwMask = 1;
    int      nTempLevel;
    int      x,y;

    // draw lines in text color
    SetBkColor(draw.hDC,  GetSysColor(COLOR_WINDOWTEXT) );

    // Draw a series of | lines for outer levels

    for ( nTempLevel = 0; nTempLevel < nLevel ; nTempLevel++)
    {
      x = nTempLevel * 2 * nBitmapWidth/XLinePos +
	  nBitmapWidth/XLinePos +
	  XBMPOFFSET;

      if ( dwConnectLevel & dwMask )
	FastRect(draw.hDC, x,rcTemp.top,1,rcTemp.bottom - rcTemp.top);

      dwMask *= 2;
    }

    // Draw the short vert line up towards the parent

    nTempLevel = nLevel-1;
    dwMask *= 2;

    x = nTempLevel * 2 * nBitmapWidth/XLinePos +
	nBitmapWidth/XLinePos +
	XBMPOFFSET;

    if ( dwConnectLevel & dwMask )
	y = rcTemp.bottom;
    else
	y = rcTemp.bottom - nLineHeight / 2;

    FastRect(draw.hDC, x, rcTemp.top, 1, y-rcTemp.top);

    // Draw short horiz bar to right
    FastRect(draw.hDC, x, rcTemp.bottom-nLineHeight/2, nBitmapWidth/XLinePos, 1);
  }

  if (nBitmapColumns==1)	// for the case of external file-bitmaps
    nColumn = 0;

  HBITMAP Oldhbm;

  if (hbm)
    // select the requested bitmap
    Oldhbm = (HBITMAP) SelectObject(hdcMem, hbm);

  // Draw the selected bitmap
  BitBlt(draw.hDC, wIndent, wTopBitmap,
	     nBitmapWidth, nBitmapHeight,
	     hdcMem,
	     nColumn * 2*(nBitmapWidth/XLinePos),
	     nRow * nBitmapHeight,
	     SRCCOPY);

  if (hbm)
    // reselect the previous bitmap
    Oldhbm = (HBITMAP) SelectObject(hdcMem, Oldhbm);

DealWithSelection:

  if (draw.itemState & ODS_SELECTED)
  {
    SetBkColor(draw.hDC, GetSysColor(COLOR_HIGHLIGHT));
    SetTextColor(draw.hDC, GetSysColor(COLOR_HIGHLIGHTTEXT));
  }
  else
  {
    SetBkColor(draw.hDC, WINDOWBKGNDCOLOR);
    SetTextColor(draw.hDC, GetSysColor(COLOR_WINDOWTEXT));
  }

  ExtTextOut(draw.hDC, rcTemp.left + 1, wTopText,
		   ETO_CLIPPED | ETO_OPAQUE,
		   &rcTemp,
		   szText, lstrlen(szText),
		   NULL);

  if (draw.itemState & ODS_FOCUS && draw.itemAction != ODA_SELECT)
  {
    ::DrawFocusRect(draw.hDC, &rcTemp);
  }
}

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
// 	class cHierDialog
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@


DEFINE_RESPONSE_TABLE1(cHierDialog, TDialog)
  EV_WM_DRAWITEM,
  EV_WM_MEASUREITEM,
END_RESPONSE_TABLE;

//------------------------------------------------------------------------
void	cHierDialog::SetupWindow()
//------------------------------------------------------------------------
// MUST NOT USE 3D effect on that type of listbox
// Code has been copied from source\owl\dialog.cpp
{
  if (GetApplication()->Ctl3dEnabled())
  {
    static BOOL FAR PASCAL(*ctl3dSubclassDlg)(HWND, WORD) = 0;

    if (!ctl3dSubclassDlg)
     (FARPROC)ctl3dSubclassDlg =
       GetApplication()->GetCtl3dModule()->GetProcAddress("Ctl3dSubclassDlg");

    if (ctl3dSubclassDlg)
    {
      // supress usage of 3D effect on listboxes
      ctl3dSubclassDlg(*this, CTL3D_ALL & ~CTL3D_LISTBOXES);
    }
  }

  SetCaption(Title);
  TWindow::SetupWindow();
}

